#!/usr/bin/python
# .+
#
# .identifier :	$Id: mainservice.py,v 1.5 2005/07/19 10:07:47 fabrizio Exp $
# .context    : program support
# .title      : program general pourpose services
# .kind	      : python source
# .author     : Fabrizio Pollastri <mxgbot@gmail.com>
# .site	      : Torino - Italy
# .creation   :	06-Feb-2004
# .copyright  :	(c) 2004 Fabrizio Pollastri
# .license    : GNU General Public License (see .copying below)
#
# .description
#
# This class definition implements a thread safe collection of general
# pourpose services for a program.
# -Command line option getting with built in options: quiet, verbose,
#  version, working directory.
# -Message logging to file, message types by increasing
#  priority: information, notice, warning, error, alarm, fatal.
# -Copying of log messages to console with verbosity control.
# -Main termination with threads termination, mailing of a session
#  report, controlled by verbose level, closing of data persistence file.
# -File versioning with max file number control.
# -Data persistence file management.
# See class methods details in the source.
#
# .copying
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# .-


# import required modules

import datetime		# time operations
import optparse		# command line option parsing
import os		# operating system support
import os.path		# file name utils
import re		# regular expressions matching searching
import shelve		# object persistence
import smtplib		# smtp mailing
import string		# string operations
import sys		# system calls
import threading	# threading support
import time		# time functions
import traceback	# traceback functions
import types		# types definitions


#### .class MainService
#

dp_root_key = '__dp_root'		# define key for data persistence root

class MainService:

  # verbosity masks for each type of log message
  FATAL_MASK		= 0x1
  ALARM_MASK		= 0x2
  ERROR_MASK		= 0x4
  WARNING_MASK		= 0x8
  NOTICE_MASK		= 0x10
  INFO_MASK		= 0x20
  ALL_MASK=FATAL_MASK|ALARM_MASK|ERROR_MASK|WARNING_MASK|NOTICE_MASK|INFO_MASK

  # verbosity levels and masks for each type of log message
  msg_verb = {	'F':(1,FATAL_MASK),
		'A':(2,ALARM_MASK),
		'E':(3,ERROR_MASK),
		'W':(4,WARNING_MASK),
		'N':(5,NOTICE_MASK),
		'I':(6,INFO_MASK)}


### the instantiator
  #
  # usage	: help usage message.
  # version	: main version string as given by rcs.
  # options	: options list by optparse.make_option.
  # args_num	: min & max of expected command line arguments
  # log_fname	: log file name.
  # max_vers	: max number of log file versions to keep.
  # mailer	: mail server host.
  # mail_dest	: mail destination address.
  # dpr		: data persistence root block.

  def __init__(	self,usage='',version='',options='',args_num=(0,10),
    log_fname=os.path.splitext(os.path.basename(sys.argv[0]))[0]+'.log',
    max_vers=5,mailer='',mail_dest='',dpr=''):

    self.program_name = os.path.basename(sys.argv[0]) # get program name
    self.session_start = time.time()	# session start time
    self.node_name = os.uname()[1]	# computer node name

    self.term_verb = self.ALL_MASK	# init terminal verbosisity level
    self.log_verb = self.ALL_MASK	# init log file verbosisity level
    self.mail_verb = self.WARNING_MASK + self.ERROR_MASK \
      + self.ALARM_MASK + self.FATAL_MASK# init mail report verbosisity level
    self.information = 0		# init total information counter
    self.notices = 0			# init total notices counter
    self.warnings = 0			# init total warnings counter
    self.errors = 0			# init total errors counter
    self.alarms = 0			# init total alarms counter
    self.fatals = 0			# init total fatals counter
    self.mailer = mailer		# set mailer name (fully qualified)
    self.mail_dest = mail_dest		# set mail destination
    self.dp = ''			# mark no open dp file
    self.dpr = ''			# mark no dp control block
    self.terminators = []		# init an empty list of terminators
    self.lf_lock = threading.RLock()	# lock for thread safe writing
   
    # create command line option parser
    version = self.program_name + ' ' + version[11:-2] # pgm name and ver num
    version += ', (c) 2004 Fabrizio Pollastri <pollastri@iriti.cnr.it>'
    usage += '\t\t[-L LOG_FNAME][-M MAIL_DEST][-q|-v VERB]\n'#---
    usage += '\t\t[-V LOG_MAIL][-W WORK_DIR]\n'#--
    usage += '\tmdgY.py -h|--version'	#- complete with built in options usage
    self.opt_parser = optparse.OptionParser(usage=usage,version=version)

    # add built in options: quiet, verbose, version, work dir
    self.opt_parser.add_option("-L","--log_file",
      action="store",type="string",dest="log_fname",
      help='set log file name to LOG_FNAME.')
      
    self.opt_parser.add_option("-M","--mail_dest",
      action="append",type="string",default=["",""],
      help='set mail destination and mailer, format <address>,<mailer> '
        + '[default=None].')
      
    self.opt_parser.add_option("-q","--quiet",
      action="store_const",const=0,dest="verb",
      help='suppress any printing to terminal.')
      
    self.opt_parser.add_option("-v","--verbose",
      action="store",type="string",dest="verb",default="2",
      help='set terminal verbosity by level or by mask. By level, integer '
        + 'with range 0-6: 0 no msg, 1 fatals (F), 2 F+alarms (A), '
	+ '3 F+A+errs (E), 4 F+A+E+warns (W), 5 F+A+E+W+notices (N), '
	+ '6 F+A+E+W+N+information (I) [default=2]. By mask, one capital '
	+ 'letter for each type of message: F for fatals, A for alarms, '
	+ 'E for errors, W for warnings, N for notices and I for information '
	+ '[default=\'FA\'].')

    self.opt_parser.add_option("-V","--log_mail_verb",
      action="store",type="string",default="",
      help='apply terminal verbosity to logging and to mail sending by mask. '
        + 'One capital letter for each application: L for logging, M for mail '
	+ '[default=None].')
      
    self.opt_parser.add_option("-W","--work_dir",
      action="store",type="string",default=".",
      help="set working directory [default=\'.\'].")

    try:				# control errors
      for option in options:		# for each given option ...
        self.opt_parser.add_option(option) # add it to parser
      # parse command line options
      self.opts, self.args = self.opt_parser.parse_args()
      os.chdir(self.opts.work_dir)	# set working directory
    except SystemExit:			# if help or version option ...
      sys.exit()			# exit
    except:				# for any other error ...
      # give fatal error msg and exit
      print 'F-'+ str(sys.exc_info()[0]) + '-' + str(sys.exc_info()[1])
      sys.exit()			# exit

    # if required, open log file
    if self.opts.log_fname:		# if log file name defined by option ...
      log_fname = self.opts.log_fname	# override log fname by arguments
    if log_fname:			# if log file name not null ...
      self.file_vers(log_fname,max_vers) # manage log file versions
      self.lf = open(log_fname,'w',1)	# open log file
    else:				# else ...
      self.lf = ''			# init log file name to empty

    # if specified by option, set mail destination and mailer
    if self.opts.mail_dest[0]:		# if destination is specified ...
      self.mail_dest = self.opts.mail_dest[0]# override it in argument
    if self.opts.mail_dest[1]:		# if mailer is specified ...
      self.mailer = self.opts.mail_dest[1]# override it in argument
    
    # convert verbose value from int or string to int
    self.term_verb = 0			# clear verbosity mask
    try:
      self.opts.verb = int(self.opts.verb)# first try it as integer
      # it is int, convert verbosity level to verbosity mask
      for msg_type in self.msg_verb.keys():# for each message type ...
        if self.msg_verb[msg_type][0] <= self.opts.verb:# if match verb level...
	  self.term_verb |= self.msg_verb[msg_type][1]# add its mask bit to mask
    except ValueError:			# if not integer ...
      try:
	for msg_type in self.opts.verb:	# for each given mask mnemonic ...
          self.term_verb |= self.msg_verb[msg_type][1]# add mask bit to mask
      except:				# if unknown mask menemonic
        self.fatal('Unknown verbosity mask mnemonic\n')# log fatal

    # if required, apply terminal verbosity to logging and mailing.
    for verb_type in self.opts.log_mail_verb:# for each given mask mnemonic ...
      if verb_type == 'L':		# if terminal verb is wanted for log...
        self.log_verb = self.term_verb	# copy it to log verbosity mask
      elif verb_type == 'M':		# if terminal verb is wanted for mail...
        self.mail_verb = self.term_verb	# copy it to mail verbosity mask
      else:				# if unknown extension menonic ...
        self.fatal('Unknown verbosity mask mnemonic\n')# log fatal

    # check option values
    if len(self.args) < args_num[0]:	# if not enough args fatal stop
      self.fatal('Less than ' + str(args_num[0]) + ' arguments')
    if len(self.args) > args_num[1]:	# if too many args fatal stop
      self.fatal('More than ' + str(args_num[1]) + ' arguments')

    # log the start message
    message = '++ ' + self.program_name + ' session start @ ' + \
      time.ctime(self.session_start)	# start msg
    self.log('',message)		# log start msg

    # if required open data persistence file and load root data block
    if dpr:				# if data persistence control block ...
      dpf_exist = os.path.isfile(dpr.fname) # check dp file existence
      try:				# control errors
        self.dp = shelve.open(dpr.fname) # open data persistence file
	if dpf_exist:			# if dp file is preexistent ...
	  self.dpr = self.dp[dp_root_key] # load dpr from file
          self.log('M','opened dp file '+dpr.fname+' and loaded dp root') # log
	else:				# if not ...
          self.dpr = dpr		# load dpr from argument
	  self.dp[dp_root_key] = dpr	# store root
          self.log('M','created dp file '+dpr.fname+' and initialized root')#log
      except:                           # if load error
        self.fatal()                    # log error msg and exit


### method: log
  # log a message to file and to terminal, if message match the given
  # verbosity level. Format of log message is <head>-HH:MM:SS-<message> .
  # If head is empty, only <message> is printed. Message can be a string
  # or a list of strings. If it is a list, <head>-HH:MM:SS- is printed only
  # in front of the first string.
  #
  # head	: log message prefix, if empty no time stamp is printed.
  # message	: message text to log.
  # verb	: verbosity level (1 fatal, 2 alarm, 3 error, 4 warning,
  #		  5 notice, 6 information).

  def log(self,head='',message='',verb=None):

    if head:				# if regular log head required ...
      shead = head + '-'		# compose short log head
      # get present log time, hh:mm:ss format.
      log_time = time.strftime('%H:%M:%S',time.localtime(time.time()))
      head += '-' + log_time + '-'	# compose complete log head
    else:				# no log head required ...
      shead = ''			# set empty short head

    # if message is not given, generate it from exception info traceback
    if len(message) == 0:		# if message is empty get std err msg
      message = [string.split(str(sys.exc_info()[-1]),'.')[-1]#--
        + ': ' + str(sys.exc_info()[1])]#- exception type and value
      message += self.trace_report()	# add traceback report

    # if verbosity not given, default to information
    if not verb:
      verb = self.INFO_MASK		# set verbosity to information
      self.information += 1		# increment total information counter

    # write message
    self.lf_lock.acquire()		# lock log file write, be thread safe
    if type(message) == types.StringType: # if message is a string ...
      if self.term_verb & verb:		# if verbosity mask match ...
        print shead + message		# copy message to console
      if self.lf and self.log_verb & verb:# if log file open and verb mask match
        self.lf.write(head + message + '\n') # write log msg
    elif type(message) == types.ListType: # if message is a list ...
      if message:			# if message no empty
        if self.term_verb & verb:	# if verbosity mask match ...
          print shead + message[0]	# copy message to console
        if self.lf and self.log_verb & verb:# if log open and verb mask match
          self.lf.write(head + message[0] + '\n') # write log msg header
      if len(message) > 1:		# if more than 1 line of message ...
        for row in message[1:]:		# each element of message list is a row
          if self.term_verb & verb:	# if verbosity mask match ...
            print '  ' + row		# copy message to console
          if self.lf and self.log_verb & verb:# if log open and verb mask match
            self.lf.write('  ' + row + '\n') # write row log msg
    else:				# if it is a wrong message type ...
      self.lf_lock.release()		# unlock log file write, by thread safe
      self.fatal('Wrong type of log message: nor list or string\n') # fatal
    self.lf_lock.release()		# unlock log file write, by thread safe


### methods: info, notice, warning, error, alarm, fatal.
  # log a given message with the respective type and continue run except for
  # fatal. Log format is the same of log method with head equal to 'I' for
  # information, 'N' for notice, 'W' for warning, 'E' for error, 'A' for alarm,
  # 'F' for fatal.
  # The fatal method do not return: it call terminate to stop execution.
  #
  # message	: message text to log.

  def info(self,message=''):		# information log
    self.information += 1		# increment total information counter
    self.log('I',message,self.INFO_MASK)# log information msg

  def notice(self,message=''):		# notice log
    self.notices += 1			# increment total notices counter
    self.log('N',message,self.NOTICE_MASK)# log notice msg

  def warning(self,message=''):		# warning log
    self.warnings += 1			# increment total warnings counter
    self.log('W',message,self.WARNING_MASK)# log warning msg

  def error(self,message=''):		# error log
    self.errors += 1			# increment total errors counter
    self.log('E',message,self.ERROR_MASK)# log error msg

  def alarm(self,message=''):		# alarm log
    self.alarms += 1			# increment total alarms counter
    self.log('A',message,self.ALARM_MASK)# log alarm msg

  def fatal(self,message=''):		# fatal log
    self.fatals += 1			# increment fatal message counter
    self.log('F',message,self.FATAL_MASK)# log fatal msg
    self.terminate()			# terminate program execution


### method: terminate
  # if warnings or errors occurred send a mail message with this kind of
  # messages extracted from the log file.
  # Then close log file and terminate program execution.

  def terminate(self):

    # call all terminators functions/methods, if any
    for (terminator,args) in self.terminators: terminator(args)

    # wait for all threads completion
    threads = threading.enumerate()	# get list of threads
    for thread in threads[1:]:		# for each thread except main thread ...
      try:
        thread.join()			# wait for its completion
      except:				# if thread terminated in the meanwhile
        continue			# goto next one

    message = self.session_report()	# build session report

    # if required, mail a report message
    if self.mailer and self.mail_dest:	# if a mail destination is defined ...
      msg_count_mask = 0		# mask of message counters not zero
      if self.information != 0:		# if there were information ...
        msg_count_mask += self.INFO_MASK# set information bit mask
      if self.notices != 0:		# if there were notices ...
        msg_count_mask += self.NOTICE_MASK# set notices bit mask
      if self.warnings != 0:		# if there were warnings ...
        msg_count_mask += self.WARNING_MASK# set warning bit mask
      if self.errors != 0:		# if there were warnings ...
        msg_count_mask += self.ERROR_MASK# set error bit mask
      if self.alarms != 0:		# if there were alarms ...
        msg_count_mask += self.ALARM_MASK# set alarm bit mask
      if self.fatals != 0:		# if there were fatals ...
        msg_count_mask += self.FATAL_MASK# set fatal bit mask

      # if at least one non zero counter match the verbosity mask, send mail.
      if msg_count_mask & self.mail_verb:
	mail = ['Subject: ' + message[0]] + message + ['----\n']
	# send message
	mailer = smtplib.SMTP(self.mailer) # instantiate smtp link
	try:
	  mailer.sendmail(self.program_name + '@' + self.node_name,
	    self.mail_dest,string.join(mail,'\n'))# send
	except:
	  self.error()			# if any send error log it

        mailer.quit()			# close connection

    # if required save dp root and close dp file
    if self.dp:				# if dp file is open ...
      self.dp[dp_root_key] = self.dpr	# save dp root
      self.dp.close()			# close db file
      self.notice('Saved dp root and closed dp file '+self.dpr.fname) # log it

    self.notice(message)		# log session report
    message = '-- ' + self.program_name + ' session end @ ' + \
      time.ctime(self.session_end)	# end log msg
    self.log('',message)		# log it
    if self.lf:				# if log file defined ...
      self.lf.close()			# close log file
    sys.exit(0)				# terminate


### method: trace_report
  # compose a traceback message report from current traceback information.
  #
  # message	: message text to log.

  def trace_report(self):

    report = ['Traceback Stack (most recent call last)'] # report header
    level = 0				# init stack level

    # for each traceback stack level ...
    for file, lineno, function, text in traceback.extract_tb(sys.exc_info()[2]):
      # write two line: the first with file, line number and function, the
      # second with the involved statement.
      report += [str(level) + ' In file "' + os.path.basename(file)
        + '" at line ' + str(lineno) + ', in function "' + function + '":',
	str(level) + '    ' + text]
      level += 1			# increment stack level

    return report			# return traceback report


### method: session_report
  # write a report at session end with the number of occurred fatals,
  # alarms, errors, warnings, notices, information messages.
  # The report is returned as a list of lines.

  def session_report(self):

    # report header
    report = self.program_name + '@' + self.node_name
   
    ## customize report for presence or not of session errors

    errors_num = self.warnings + self.errors + self.alarms + self.fatals

    if errors_num == 0:			# if not errors ...
      report = [report + ' session OK!']
    else:				# if errors ...
      report = [report + ' session WARNINGS/ERRORS/ALARMS/FATAL!']

    ## write session start, end and elapsed times

    report += [string.ljust('Session start',15) + ': '
      + time.ctime(self.session_start)]
    self.session_end = time.time()	# set session end time
    report += [string.ljust('Session end',15)+': '+time.ctime(self.session_end)]
    # elapsed time in secs
    elapsed = datetime.datetime.fromtimestamp(self.session_end) \
      - datetime.datetime.fromtimestamp(self.session_start)
    report += [string.ljust('Elapsed time',15) + ': ' + str(elapsed)]
    
    ## write all non zero message counters values to report

    counters_names = [			# list of all counters and their names
      (self.information,'information'),
      (self.notices + 1,'notices'),
      (self.warnings,'warnings'),
      (self.errors,'errors'),
      (self.alarms,'alarms'),
      (self.fatals,'fatals')]
    
    zero_msg = ''			# init an empty zero msg

    for counter, name in counters_names:# for each message counter ...
      if counter == 0:			# if no messages of this kind ...
        zero_msg += ', no ' + name	# add "no <kind>" to zero message string
      else:				# if there are message of this kind ...
        name = string.ljust(name,15) + ':'# counter name with fixed string len
        value = string.rjust(str(counter),5)# the same for its value
        report += [name + value]	# add it to report

    if zero_msg:			# if non empty ...
      report += [zero_msg[2:] + '.']	# add zero counter message
    
    return report			# return session report
	

### method: file_vers
  # this method backups older versions of a given file (GF) into files
  # whose names have the form <GF>.<version number (VN)> . The older the
  # version, the smaller the version number. VN starts from 1. 
  # 1) if GF does not exist, then do nothing.
  # 2) if GF exists and <GF>.<VN> do not exist, then GF is moved to <GF>.1 .
  # 3) if GF and <GF>.<PN> exist, then GF is moved to <GF>.<max(PN)+1> .
  # This method can also keep only a given number of the most recent versions
  # and delete the older versions exceeding the given number.
  #
  # fname	: file name to manage version.
  # keep	: number of versions to keep.

  def file_vers(self,fname='',keep=10):

    if os.path.exists(fname):		# if file exists ...
      path, name = os.path.split(fname)	# separate path and name
      if not path:			# if no path ...
        path = '.'			# assume current dir
      files = os.listdir(path)		# list all dir files
      vfile_re = re.compile(name + '.[1-9][0-9]*') # versioned file names regex
      # select only files ending with version number, if any
      vfiles = [vfile for vfile in files if vfile_re.match(vfile)]
      
      if vfiles:			# if exist one or more older versions
        # get all existing number of file versions
        ver_nums = [string.split(vfile,'.')[-1] for vfile in vfiles]
	ver_nums = map(int,ver_nums)	# convert string to integer
	new_ver_num = max(ver_nums) + 1	# compute new version number
	os.rename(fname,fname+'.'+str(new_ver_num)) # move fname to new version

	if keep and keep <= len(ver_nums): # if too many versions ...
	  ver_nums.sort()		# ascending sort
	  ver_nums = map(str,ver_nums)	# convert integer to string

	  # for each version exceeding keep num ...
	  for ver_num in ver_nums[:len(ver_nums) - keep + 1]:
	    os.remove(fname + '.' + ver_num) # delete it

      else:				# no older version so ...
        os.rename(fname,fname + '.1')	# create first
#
#### .class MainService
