#!/usr/bin/python
# .+
#
# .identifier :	$Id: textedit.py,v 1.2 2004/10/19 13:38:08 fabrizio Exp $
# .context    : text editing
# .title      : text buffer editor
# .kind	      : python source
# .author     : Fabrizio Pollastri <mxgbot@gmail.com>
# .site	      : Torino - Italy
# .creation   :	06-Feb-2004
# .copyright  :	(c) 2004 Fabrizio Pollastri
# .license    : GNU General Public License (see .copying below)
#
# .description
#
# This class can edit a text stored in a single original buffer
# producing the result into an edited buffer. Editing commands must be given
# sequentially ordered by increasing position of application into text.
# Available commands
# -Read a text file into original buffer.
# -Write edited buffer into a text file.
# -Delete a string of text at current working position.
# -Delete current line. 
# -Insert a string of text at current working position.
# -Join current line with following one.
# -Flush all text buffers: original and edited.
# See class methods for details.
#
# .copying
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# .-


# import required modules

import string		# string operations


#### .class TextEdit
# 

class TextEdit:

### the instantiator
  #
  # text	: a text to fill the original ext buffer.

  def __init__(self,text='',fname=''):

    self.fname = fname			# text file name
    self.original = text		# original text buffer
    self.edited = ''			# edited text buffer
    self.orig_pos = 0			# original text work position
    self.orig_rd_pos = 0		# original text read pointer


### method: read
  # read a text file into original buffer.
  #
  # fname	: name of text file to read from.

  def read(self,fname=''):

    if fname:				# if fname is given ...
      self.fname = fname		# store it, else keep previous
    f = open(self.fname,'r')		# open file
    self.original = f.read()		# get file text
    f.close()				# close file
    self.edited = ''			# clear edited text buffer
    self.orig_pos = 0			# reset original work position
    self.orig_rd_pos = 0		# reset original read pointer


### method: write
  # write edited buffer into a text file.
  #
  # fname	: file name to write in.

  def write(self,fname=''):

    # if there is a final part of original not yet copied to edited, copy it.
    self.edited += self.original[self.orig_rd_pos:]
    if fname:				# if fname is given ...
      self.fname = fname		# store it, else keep previous
    f = open(self.fname,'w')		# open file
    f.write(self.edited)		# write file text
    f.close()				# close file


### method: delete
  # delete a string of text at current working position.
  #
  # length	: number of chars to delete at original pointer.

  def delete(self,length):

    # if text to copy from orig to edited, copy it.
    self.edited += self.original[self.orig_rd_pos:self.orig_pos]
    self.orig_pos += length		# work pointer forward by del length
    self.orig_rd_pos = self.orig_pos	# the same for read pointer


### method: del_line
  # delete current line. 

  def del_line(self):

    # find start of line
    line_start = string.rfind(self.original,'\n',0,self.orig_pos + 1)
    if line_start == -1:		# if not found ...
      line_start = 0			# line starts at buffer start
    # find end of line
    line_end = string.find(self.original,'\n',self.orig_pos + 1)
    if line_end == -1:			# if not found ...
      line_end = len(self.original)	# line ends at buffer end
    self.delete(line_end - line_start + 1) # delete line


### method: insert
  # insert a string of text at current working position.

  def insert(self,text):

    # if text to copy from orig to edited, copy it.
    self.edited += self.original[self.orig_rd_pos:self.orig_pos]
    self.orig_rd_pos = self.orig_pos	# update read pointer
    self.edited += text			# append text to edited buffer


### method: join
  # join current line with following one.

  def join(self):

    # find end of current line
    line_end = string.find(self.original,'\n',self.orig_pos)
    if line_end == -1:			# if not found ...
      line_end = len(self.original)	# line ends at buffer end
    # copy original to edited until current line end
    self.edited += self.original[self.orig_rd_pos:line_end]
    line_end = string.find(self.original,'\n',line_end) # find end of next line
    if line_end == -1:			# if not found ...
      line_end = len(self.original)	# line ends at buffer end
    # copy original to edited until next line end
    self.edited += self.original[self.orig_rd_pos:line_end]
    self.orig_pos = line_end		# update work position to next line end
    self.orig_rd_pos = self.orig_pos	# update read pointer


### method: flush
  # flush (empty) all text buffers: original and edited.

  def flush(self):

    self.original = ''			# flush original ext buffer
    self.edited = ''			# flush edited text buffer
#
#### .class TextEdit
